from functools import lru_cache
from typing import Dict, List

from nagra import execute
from nagra.schema import Schema
from nagra.delete import Delete
from nagra.select import Select
from nagra.upsert import Upsert
from nagra.transaction import Transaction


_SQLITE_TYPE_MAP = {
    "varchar": "TEXT",
    "int": "INTEGER",
    "float": "FLOAT",
    "timestamp": "DATETIME",
}



class Table:
    def __init__(
        self,
        name: str,
        columns: Dict,
        natural_key: List = None,
        foreign_keys: Dict = None,
        not_null: List = None,
    ):
        self.name = name
        self.columns = columns
        self.natural_key = natural_key or list(columns)
        self.foreign_keys = foreign_keys or {}
        self.not_null = set(self.natural_key) | set(not_null or [])
        self.schema = Schema.default
        self.schema.add(self.name, self)

    @classmethod
    def get(self, name):
        """
        Shortcut method to Schema.default().get()
        """
        return Schema.default.get(name)

    def select(self, *columns, where=None):
        if not columns:
            columns = list(self.default_columns())
        slct = Select(self, *columns)
        if where:
            slct.where(where)
        return slct

    def join(self, env):
        for prefix, alias in env.refs.items():
            # Find alias of previous join in the chain
            *head, tail = prefix
            prev_table = env.refs[tuple(head)] if head else self.name
            # Identify last table & column of the chain
            ftable, join_col = self.join_on(prefix)
            yield (ftable.name, alias, prev_table, join_col)

    def default_columns(self, nk_only=False):
        columns = self.natural_key if nk_only else self.columns
        for column in columns:
            if column not in self.foreign_keys:
                yield column
                continue

            ftable = self.schema.get(self.foreign_keys[column])
            yield from (f"{column}.{k}" for k in ftable.default_columns(nk_only=True))

    @lru_cache
    def join_on(self, path):
        if len(path) == 1:
            join_col = path[0]
            table = self
            ftable = self.schema.get(self.foreign_keys[join_col])
            return ftable, join_col
        ftable, _ = self.join_on(path[:-1])
        table = ftable
        join_col = path[-1]
        ftable = self.schema.get(table.foreign_keys[join_col])
        return ftable, join_col

    def delete(self, where=None):
        delete = Delete(self)
        if where:
            delete.where(where)
        return delete

    def upsert(self, *columns, lenient=None):
        """
        Create an upsert object based on the given columns, if
        lenient is set, foreign keys wont be enforced on the given
        columns even if a value is passed on the subsequent execute or
        executemany. Example:

        >>> upsert = Table.get("comment").upsert("body", "blog_post.title", lenient=["blog_post"])
        >>> upsert.execute(("Nice post!", "A post title that will change soon."))

        If lenient is set to True all foreign keys will be treated as such.
        """
        return Upsert(self, *columns, lenient=lenient)

    def insert(self, *columns, lenient=None):
        """
        Provide an insert-only statement (won't raise error if
        record already exists). See `Table.upsert` for `lenient` role.
        """
        return Upsert(self, *columns, lenient=None).insert_only()

    def suggest(self, column, like=None):
        """
        Return a iterator over the possible values of columns. Use
        like in a where condition if given.
        """
        if "." not in column:
            raise NotImplementedError("TODO")
        local_col, remote_col = column.split(".", 1)
        cond = []
        if like:
            cond.append("(like " + remote_col + " {})")

        ftable = Schema.get(self.foreign_keys[local_col])
        select = ftable.select(remote_col).where(*cond).groupby(remote_col).orderby(remote_col)
        if like:
            cur = execute(select.stm(), (like,))
            return (x for x, in cur)
        return (x for x, in select)

    def ctypes(self):
        if Transaction.flavor == "sqlite":
            return {
                c: _SQLITE_TYPE_MAP.get(d, d) for c, d in self.columns.items()
            }
        return self.columns



from collections import defaultdict
from itertools import islice

from nagra import Statement, executemany, execute, Transaction
from nagra.transaction import ExecMany
from nagra.sexpr import AST
from nagra.schema import Schema
from nagra.exceptions import UnresolvedFK
from nagra.utils import logger


class Upsert:
    def __init__(self, table, *columns, lenient=None):
        self.table = table
        self.columns = columns
        self.columns_ast = [AST.parse(c) for c in columns]
        self.groups, self.resolve_stm = self.prepare()
        self._insert_only = False
        self.lenient = lenient or []

    def stm(self):
        conflict_key = ["id"] if "id" in self.groups else self.table.natural_key
        columns = self.groups
        do_update = False if self._insert_only else len(columns) > len(conflict_key)
        stm = Statement(
            "upsert",
            table=self.table.name,
            columns=columns,
            conflict_key=conflict_key,
            do_update=do_update,
        )

        # TODO add returning id to upsert sql and re-query those ids to make sure acl are respected

        return stm()

    def insert_only(self):
        self._insert_only = True
        return self

    def prepare(self):
        """
        Organise columns in groups and prepare statement to
        resolve fk based on columns expressions
        """
        groups = defaultdict(list)
        for col in self.columns:
            if "." in col:
                head, tail = col.split(".", 1)
                groups[head].append(tail)
            else:
                groups[col] = None

        resolve_stm = {}
        for col, to_select in groups.items():
            if not to_select:
                continue
            cond = ["(= %s {})" % c for c in to_select]
            ftable = Schema.get(self.table.foreign_keys[col])
            select = ftable.select("id").where(*cond)
            resolve_stm[col] = select.stm()
        return groups, resolve_stm

    def execute(self, *values):
        self.executemany([values])

    def executemany(self, records):
        # Transform list of records into a dataframe-like dict
        value_df = dict(zip(self.columns, zip(*records)))
        arg_df = {}
        for col, to_select in self.groups.items():
            if to_select:
                values = list(zip(*(value_df[f"{col}.{s}"] for s in to_select)))
                arg_df[col] = self._resolve(col, values)
            else:
                arg_df[col] = value_df[col]

        args = zip(*(arg_df[c] for c in self.groups))
        # Work by chunks
        stm = self.stm()
        while True:
            chunk = list(islice(args, 1000))
            if not chunk:
                break
            if Transaction.flavor == "sqlite":
                for item in chunk:
                    execute(stm, item)
            else:
                executemany(stm, chunk)

    def _resolve(self, col, values):
        stm = self.resolve_stm[col]
        exm = ExecMany(stm, values)
        for res, vals in zip(exm, values):
            if res is not None:
                yield res[0]
            elif any(v is None for v in vals):
                # One of the values is not given
                yield None
            elif self.lenient is True or col in self.lenient:
                msg = "Value '%s' not found for foreign key column '%s' of table %s"
                logger.info(msg, vals, col, self.table)
                yield None
            else:
                raise UnresolvedFK(f"Unable to resolve '{vals}' (for column '{col}')")

    def __call__(self, records):
        return self.executemany(records)



import toml
from jinja2 import Template

from nagra import Statement, execute, Transaction


D2_TPL = """
{{table.name}}: {
  shape: sql_table
  {%- for col, dt in table.columns.items() %}
  {{col}}: {{dt}}
  {%- endfor %}
}
{%- for col, f_table in table.foreign_keys.items() %}
{{table.name}}.{{col}} -> {{f_table}}.id : {{col}}
{%- endfor -%}
"""



class Schema:
    _default = None

    def __init__(self):
        self.tables = {}

    def add(self, name, table):
        # if name in self.tables: TODO
        #     raise RuntimeError(f"Table {name} already in schema!")
        self.tables[name] = table

    @classmethod
    def get(cls, name):
        return cls.default.tables[name]

    @classmethod
    @property
    def default(cls):
        if not cls._default:
            cls._default = Schema()
        return cls._default

    def setup_statements(self):
        # Create tables
        for name, table in self.tables.items():
            ctypes = {c: t for c, t in table.ctypes().items() if c not in table.foreign_keys}
            stmt = Statement(
                "create_table",
                table=name,
                columns=ctypes,
                not_null=table.not_null,
            )
            yield stmt()

        # Add foreign keys, we need to find existing ones first
        if Transaction.flavor == "sqlite":
            fks = set()
            for name in self.tables:
                stmt = Statement("select_foreign_keys").name(name)
                rows = execute(stmt())
                fks.update(
                    (name, column, other_table)
                    for _, _, other_table, column, *_ in rows
                )
        else:
            stmt = Statement("select_foreign_keys")
            fks = set(execute(stmt()))

        for name, table in self.tables.items():
            ctypes = table.ctypes()
            for column, other_table in table.foreign_keys.items():
                if (name, column, other_table) in fks:
                    continue

                # XXX this loop will fail on composite keys (but do we need it?)
                stmt = Statement(
                    "add_foreign_key",
                    table=name,
                    column=column,
                    col_def=ctypes[column],
                    other_table=other_table,
                    not_nulld=column in table.not_null,
                )
                yield stmt()

        # Add index on natural keys
        for name, table in self.tables.items():
            stmt = Statement(
                "create_unique_index",
                table=name,
                natural_key=table.natural_key,
            )
            yield stmt()

    def setup(self):
        for stm in self.setup_statements():
            execute(stm)

    def drop(self):
        for name in self.tables:
            stmt = Statement("drop_table", name=name)
            execute(stmt())

    def generate_d2(self):
        tpl = Template(D2_TPL)
        tables = self.tables.values()
        res = "\n".join(tpl.render(table=t) for t in tables)
        return res


def load_schema(toml_file, create_tables=False):
    # Late import to avoid circual deps (should put this code in a
    # "misc" sumodule)
    from nagra.table import  Table

    # load table definitions
    tables = toml.loads(open(toml_file).read())
    for name, info in tables.items():
        Table(name, **info)
    if create_tables:
        Schema.default.setup()



import os
import bisect
import logging
import sys
from contextlib import contextmanager
from pathlib import Path
from time import perf_counter


from jinja2 import FileSystemLoader, Environment

HERE = Path(__file__).parent
DEFAULT_FLAVOR = "postgresql"  # TODO use a "settings" dataclass


fmt = "%(levelname)s:%(asctime).19s: %(message)s"
logging.basicConfig(format=fmt)
logger = logging.getLogger("nagra")
if os.environ.get("NAGRA_DEBUG"):
    logger.setLevel("DEBUG")
    logger.debug("Log level set to debug")


def autoquote(x):
    if x.startswith('"'):
        return x
    return f'"{x}"'


# Setup jinja env
jinja_env = Environment(loader=FileSystemLoader(HERE / "template"))
jinja_env.filters["autoquote"] = autoquote


def template(name, ext="sql"):
    return jinja_env.get_template(f"{name}.{ext}")


def strip_lines(stmt):
    for r in stmt.splitlines():
        line = r.strip()
        if line:
            yield line


def pretty_nb(number):
    prefixes = "yzafpnum_kMGTPEZY"
    factors = [1000**i for i in range(-8, 8)]
    if number == 0:
        return 0
    if number < 0:
        return "-" + pretty_nb(-number)
    idx = bisect.bisect_right(factors, number) - 1
    prefix = prefixes[idx]
    return "%.2f%s" % (number / factors[idx], "" if prefix == "_" else prefix)


@contextmanager
def timeit(title=""):
    start = perf_counter()
    yield
    delta = perf_counter() - start
    print(title, pretty_nb(delta) + "s", file=sys.stderr)



from nagra import execute, executemany


class Executable:
    def execute(self, *args):
        return execute(self.stm(), args)

    def executemany(self, args):
        return executemany(self.stm(), args)

    def __iter__(self):
        return iter(self.execute())



import re
import dataclasses

from nagra import Statement
from nagra.mixin import Executable
from nagra.sexpr import Env, AST, AggToken


RE_VALID_IDENTIFIER = re.compile(r"\W|^(?=\d)")

def clean_col(name):
    return RE_VALID_IDENTIFIER.sub('_', name)


class Select(Executable):
    def __init__(self, table, *columns):
        self.table = table
        self.env = Env(table)
        self.columns = columns
        self.columns_ast = [AST.parse(c) for c in columns]
        self.query_columns = [a.eval(self.env) for a in self.columns_ast]
        self.where_asts = []
        self.where_conditions = []
        self._offset = None
        self._limit = None
        self.groupby_ast = None
        self.order_ast = []
        self.order_directions = []
        super().__init__()

    def where(self, *conditions):
        asts = [AST.parse(cond) for cond in conditions]
        self.where_asts += asts
        self.where_conditions += [ast.eval(self.env) for ast in asts]
        return self

    def offset(self, value):
        self._offset = value
        return self

    def limit(self, value):
        self._limit = value
        return self

    def groupby(self, *groups):
        self.groupby_ast = [AST.parse(g) for g in groups]
        return self

    def to_dataclass(self, *aliases):
        # TODO return nullable union type if a column is not required
        return dataclasses.make_dataclass(
            self.table.name,
            fields=[(clean_col(c), d) for c, d in self.dtypes(*aliases)]
        )

    def dtypes(self, *aliases):
        fields = []
        if aliases:
            assert len(aliases) == len(self.columns)
            col_names = aliases
        else:
            col_names = self.columns
        for col_name, col_ast in zip(col_names, self.columns_ast):
            col_type = col_ast.eval_type(self.env)
            fields.append((col_name, col_type))
        return fields

    def orderby(self, *orders):
        expressions = []
        directions = []
        for o in orders:
            if isinstance(o, tuple):
                expression = o[0]
                direction = o[1]
            else:
                expression = o
                direction = "asc"

        if isinstance(expression, int):
            expression = self.columns[expression]
        expressions.append(expression)
        directions.append(direction)

        self.order_ast = [AST.parse(e) for e in expressions]
        self.order_directions = directions
        return self

    def infer_groupby(self):
        # Detect aggregates
        for a in self.columns_ast:
            if any(isinstance(tk, AggToken) for tk in a.chain()):
                break
        else:
            # No aggregate found
            return []

        # Collect non-aggregates
        groupby_ast = []
        for a in self.columns_ast:
            if any(isinstance(tk, AggToken) for tk in a.chain()):
                continue
            groupby_ast.append(a)
        return groupby_ast

    def stm(self):
        groupby_ast = self.groupby_ast or self.infer_groupby()
        groupby = [a.eval(self.env) for a in groupby_ast]

        orderby = [a.eval(self.env) + f" {d}" for a, d in zip(
            self.order_ast,
            self.order_directions,
        )]
        joins = self.table.join(self.env)
        stm = Statement(
            "select",
            table=self.table.name,
            columns=self.query_columns,
            joins=joins,
            conditions=self.where_conditions,
            limit=self._limit,
            offset=self._offset,
            groupby=groupby,
            orderby=orderby,
        )
        return stm()

    def to_pandas(self, rows, aliases=[]):
        """
        Convert the rows into columns and return a df with the
        proper column types, and the given aliases as column names.
        """
        from pandas import DataFrame, Series
        names, dtypes = zip(*(self.dtypes(*aliases)))
        by_col = zip(*rows)
        df = DataFrame()
        for name, dt, col in zip(names, dtypes, by_col):
            # FIXME Series(col, dtype=dt) fail on json cols!
            srs = Series(col)
            if dt == int:
                # Make sure we have no nan for int columns
                srs = srs.fillna(0)
            df[name] = srs.astype(dt)
        return df

    def to_dict(self):
        columns = [f.name for f in dataclasses.fields(self.to_dataclass())]
        return [dict(zip(columns, record)) for record in self]



from setuptools import setup


setup(
    name="Nagra",
    version="0.0.2",
    description="Query databases without models",
    install_requires=[
        "jinja2",
        "tabulate",
        "toml",
    ],
    packages=[
        "nagra",
    ],
    extras_require={
        "test": ["pytest"],
    },
    package_data={"nagra": ["template/*/*sql"]},  # TODO test this works!
    entry_points={
        "console_scripts": [
            "nagra = nagra.cli:run",
        ],
    },
)



from itertools import chain
import argparse
import os

from tabulate import tabulate

from nagra import Transaction, Table, Schema
from nagra import load_schema


def select(args):
    select = Table.get(args.table).select(*args.columns)
    if args.where:
        where = chain.from_iterable(args.where)
        select.where(*where)
    if args.limit:
        select = select.limit(args.limit)
    if args.orderby:
        select = select.orderby(args.orderby)
    rows = list(select.execute())
    headers = [d[0] for d in select.dtypes()]
    print(tabulate(rows, headers))


def delete(args):
    delete = Table.get(args.table).delete()
    where = chain.from_iterable(args.where)
    delete.where(*where)
    delete.execute()


def schema(args):
    sch = Schema.default
    if args.d2:
        print(sch.generate_d2())
        return

    rows = []
    for name in sch.tables:
        rows.append([name])
    headers = ["table"]
    print(tabulate(rows, headers))


def run():
    # top-level parser
    parser = argparse.ArgumentParser(
        prog="nagra",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    default_db = os.environ.get('NAGRA_DB')
    default_schema = os.environ.get('NAGRA_SCHEMA')
    parser.add_argument(
        "--db",
        "-d",
        default=default_db,
        help=f"DB uri, (default: {default_db})",
    )
    parser.add_argument(
        "--schema",
        "-s",
        default=default_schema,
        help=f"DB schema, (default: {default_schema})",
    )
    subparsers = parser.add_subparsers(dest="command")

    parser_select = subparsers.add_parser("select")
    parser_select.add_argument("table")
    parser_select.add_argument("columns", nargs="*")
    parser_select.add_argument("--where", "-W", type=str, action='append', nargs="*")
    parser_select.add_argument("--limit", "-L", type=int)
    parser_select.add_argument("--orderby", "-O", help="Order by given column")
    parser_select.set_defaults(func=select)

    parser_delete = subparsers.add_parser("delete")
    parser_delete.add_argument("table")
    parser_delete.add_argument("--where", "-W", type=str, action='append', nargs="*")
    parser_delete.set_defaults(func=delete)

    parser_schema = subparsers.add_parser("schema")
    parser_schema.add_argument("--d2", action="store_true",
                               help="Generate d2 file")
    parser_schema.set_defaults(func=schema)

    # Parse args
    args = parser.parse_args()
    if not args.command:
        parser.print_help()
        return

    try:
        with Transaction(args.db):
            load_schema(args.schema)
            args.func(args)
    except (BrokenPipeError, KeyboardInterrupt):
        pass



from nagra import Statement
from nagra.mixin import Executable
from nagra.sexpr import Env, AST


class Delete(Executable):
    def __init__(self, table):
        self.table = table
        self.env = Env(table)
        self.where_asts = []
        self.where_conditions = []
        super().__init__()

    def where(self, *conditions):
        asts = [AST.parse(cond) for cond in conditions]
        self.where_asts += asts
        self.where_conditions += [ast.eval(self.env) for ast in asts]
        return self

    def stm(self):
        joins = list(self.table.join(self.env))
        stm = Statement(
            "delete-with-join" if joins else "delete",
            table=self.table.name,
            joins=joins,
            conditions=self.where_conditions,
        )
        return stm()

    def args(self):
        res = []
        for ast in self.where_asts:
            res += ast.get_args()
        return res

    def __call__(self):
        return self.execute()



"""
The `AST` (abstract syntax tree) implement parsing and evaluation
of [s-expressions](https://en.wikipedia.org/wiki/S-expression).

Example:

``` python-console
>>> from nagra.sexpr import AST
>>> AST.parse('(+ 1 1)')
<nagra.sexpr.AST object at 0x7f1bcc5b2fd0>
>>> ast = AST.parse('(+ 1 1)')
>>> ast.eval()
2
```

The `eval` method accepts an `env` parameter, a dictionary used to
evaluate non-litteral tokens:

``` python-console
>>> ast = AST.parse('(+ 1 x)')
>>> ast.eval()
Traceback (most recent call last):
   ...
ValueError: Unexpected token: "x"
>>> ast.eval({'x': 2})
3
```

"""

import shlex
from datetime import datetime

from nagra import Transaction

__all__ = ["AST"]


UNSET = object()


def list_to_dict(*items):
    it = iter(items)
    return dict(zip(it, it))


class KWargs:
    def __init__(self, *items):
        self.value = list_to_dict(*items)

    def __repr__(self):
        return f'<KWargs "{self.value}">'


class Env:
    def __init__(self, table):
        self.table = table
        self.refs = {}

    def add_ref(self, path):
        *head, name, tail = path
        prefix = tuple([*head, name])
        table_alias = self.refs.get(prefix)
        if not table_alias:
            if len(prefix) >= 2:
                self.add_ref(prefix)
            table_alias = f"{name}_{len(self.refs)}"
            self.refs[prefix] = table_alias
        return f'"{table_alias}"."{tail}"'

    def __repr__(self):
        content = repr(self.refs)
        return f"<Env {self.table.name} {content}>"


class Alias:
    """
    Simple wrapper that combine a value and an alias name
    """

    def __init__(self, value, name):
        self.value = value
        self.name = name


def tokenize(expr):
    lexer = shlex.shlex(expr)
    lexer.wordchars += ".!=<>:{}-"
    for i in lexer:
        yield Token.from_value(i)


def scan(tokens, end_tk=")"):
    res = []
    for tk in tokens:
        if tk.value == end_tk:
            return res
        elif tk.value == "(":
            res.append(scan(tokens))
        else:
            res.append(tk)

    tail = next(tokens, None)
    if tail:
        raise ValueError(f'Unexpected token: "{tail.value}"')
    return res


class AST:
    builtins = {
        "!=": "{} != {}".format,
        "<": "{} < {}".format,
        "<=": "{} <= {}".format,
        "=": "{} = {}".format,
        ">": "{} > {}".format,
        ">=": "{} >= {}".format,
        "and": lambda *x: " AND ".join(x),
        "or": lambda *x: " OR ".join(x),
        "not": "NOT {}".format,
        "is": "{} is {}".format,
        "in": lambda x, *ys: f"{x} in (%s)" % ", ".join(ys),
        "null": lambda: "NULL",
        "true": lambda: "true",
        "false": lambda: "false",
        "like": "{} LIKE {}".format,
        "ilike": "{} ILIKE {}".format,
        "+" :lambda *xs: (' + '.join("{}" for _ in xs)).format(*xs),
        "-": lambda *xs: (' - '.join("{}" for _ in xs)).format(*xs),
        "*": lambda *xs: (' * '.join("{}" for _ in xs)).format(*xs),
        "/": lambda *xs: (' / '.join("{}" for _ in xs)).format(*xs),
    }

    aggregates = {
        "min": "min({})".format,
        "max": "max({})".format,
        "sum": "sum({})".format,
        "avg": "avg({})".format,
        "every": "every({})".format,
        "count": lambda x="*": f"count({x})",
    }

    def __init__(self, tokens):
        # Auto-wrap sublist into AST
        self.tokens = [
            tk if isinstance(tk, Token) else AST(tk)
            for tk in tokens
        ]

    @classmethod
    def parse(cls, expr):
        res = tokenize(expr)
        tokens = scan(res)[0]
        if isinstance(tokens, Token):
            tokens = [tokens]
        return AST(tokens)

    def chain(self):
        for tk in self.tokens:
            if isinstance (tk, Token):
                yield tk
            else:
                yield from tk.chain()

    def _eval(self, env, top=False):
        head, tail = self.tokens[0], self.tokens[1:]
        args = [tk._eval(env) for tk in tail]
        res = head._eval(env, *args)
        return res if top else "({})".format(res)

    def eval(self, env):
        return self._eval(env, top=True)

    def relations(self):
        for tk in self.chain():
            if tk.is_relation():
                yield tk.value

    def _eval_type(self, env):
        head, tail = self.tokens[0], self.tokens[1:]
        args = [tk._eval_type(env) for tk in tail]
        res = head._eval_type(env, *args)
        return res

    def eval_type(self, env):
        return self._eval_type(env)


    def get_args(self):
        """
        Return token that should be treated as query arguments
        """
        args = list((tk.get_arg() for tk in self.chain()))
        return [a for a in args if a]

    def is_aggregate(self):
        for tk in self.chain():
            if tk.is_aggregate():
                return True
        return False


class Token:
    def __init__(self, value):
        self.value = value

    def is_relation(self):
        return False

    @staticmethod
    def from_value(value):
        if value in AST.builtins:
            return BuiltinToken(value)
        if value in AST.aggregates:
            return AggToken(value)
        if (value[0], value[-1]) == ("{", "}"):
            return ParamToken(value)
        try:
            if '.' in value:
                value = float(value)
                return FloatToken(value)
            else:
                value = int(value)
                return IntToken(value)
        except ValueError:
            pass
        return StrToken(value) if value[0] in "\"'" else VarToken(value)

    def __repr__(self):
        cls = self.__class__.__name__
        return f"<{cls} {self.value}>"

    def get_arg(self):
        return None

    def _eval(self, env, *args):
        return None


class ParamToken(Token):
    "Parameterized Token"

    def __init__(self, value):
        # Remove braces
        self.value = value[1:-1]

    def _eval(self, env, *args):
        placeholder = "%s" if Transaction.flavor == "postgresql" else "?"
        return placeholder


class LitToken(Token):
    "Litteral Token"

    def _eval_type(self, env):
        return self._type


    def _eval(self, env, *args):
        return self.value

class FloatToken(LitToken):
    "Float Token"
    _type = float


class IntToken(LitToken):
    "Integer Token"
    _type = int


class StrToken(LitToken):
    "String Token"

    def __init__(self, value):
        # Remove quotes
        self.value = value[1:-1]

    def _eval_type(self, env):
        return str

    def _eval(self, env, *args):
        return f"'{self.value}'"


class VarToken(Token):
    def __init__(self, value):
        self.join_alias = None
        super().__init__(value)

    def is_relation(self):
        return "." in self.value

    def _eval(self, env, *args):
        if self.is_relation():
            self.join_alias = env.add_ref(self.value.split("."))
            return self.join_alias
        return '"{}"."{}"'.format(env.table.name, self.value)

    def _eval_type(self, env):
        if self.value == "id":
            # id is implicit on Table
            return int
        # TODO handle paramtoken here?
        if self.is_relation():
            *head, tail = self.value.split(".")
            ftable, _ = env.table.join_on(tuple(head))
            col_type = ftable.columns[tail]
        else:
            col_type = env.table.columns[self.value]
        match col_type:
            # TODO validate type names int Table.__init__
            case "int" | "bigint":
                return int
            case "varchar":
                return str
            case "float":
                return float
            case "timestamp":
                return datetime
            case "bool":
                return bool
            case "json":
                return str
            case _:
                msg = f"Columns of type {col_type} not supported (for {self.value})"
                raise NotImplementedError(msg)


class OpToken(Token):
    ops = AST.builtins

    def _eval(self, env, *args):
        op = self.ops[self.value]
        return op(*args)


class BuiltinToken(OpToken):
    num_like = "+-*/"
    bool_like = (
        "!=", "<", ">", ">=", "<=", "=",
        "and", "or", "not", "is", "like", "ilike",
    )

    def _eval_type(self, env, *operands):
        # FIXME, probably too basic
        if self.value in self.num_like:
            if any(op == float for op in operands):
                return float
            return int
        elif self.value in self.bool_like:
            return bool
        else:
            return str

class AggToken(OpToken):
    ops = AST.aggregates

    num_like = ["sum", "avg"]
    bool_like = ["every"]

    def _eval_type(self, env, *operands):
        if self.value == "count":
            return int
        if self.value in self.num_like:
            assert operands[0] in (float, int)
            return operands[0]
        if self.value in self.bool_like:
            assert operands[0] == bool
            return bool
        else:
            return operands[0]



import sqlite3
import threading

from nagra.utils import logger, DEFAULT_FLAVOR
from nagra.exceptions import NoActiveTransaction


class Transaction:

    _local = threading.local()
    _local.current_transaction = []

    def __init__(self, dsn, rollback=False):
        if dsn.startswith("postgresql://"):
            import psycopg
            # TODO use Connection Pool
            self.flavor = "postgresql"
            self.connection = psycopg.connect(dsn)
        elif dsn.startswith("sqlite://"):
            self.flavor = "sqlite"
            filename = dsn[9:]
            self.connection = sqlite3.connect(filename)
        elif dsn.startswith("duckdb://"):
            import duckdb
            self.flavor = "duckdb"
            filename = dsn[9:]
            self.connection = duckdb.connect(filename)
            self.connection.begin()
        else:
            raise ValueError(f"Invalid dsn string: {dsn}")
        self.cursor = self.connection.cursor()
        self.rollback = rollback

    @classmethod
    @property
    def current(cls):
        try:
            return cls._local.current_transaction[-1]
        except IndexError:
            raise NoActiveTransaction()

    @classmethod
    @property
    def flavor(cls):
        try:
            return Transaction.current and Transaction.current.flavor or DEFAULT_FLAVOR
        except NoActiveTransaction:
            return DEFAULT_FLAVOR

    @classmethod
    def execute(cls, stmt, args=tuple()):
        logger.debug(stmt)
        transaction = cls.current
        cursor = transaction.cursor
        cursor.execute(stmt, args)
        if transaction.flavor == "duckdb":
            return yield_from_cursor(cursor)
        else:
            return cursor


    @classmethod
    def executemany(cls, stmt, args=None):
        logger.debug(stmt)
        transaction = cls.current
        cursor = transaction.cursor
        cursor.executemany(stmt, args)
        if transaction.flavor == "duckdb":
            return yield_from_cursor(cursor)
        else:
            return cursor

    def __enter__(self):
        if not hasattr(self._local, "current_transaction"):
            self._local.current_transaction = []
        self._local.current_transaction.append(self)
        return self.cursor

    def __exit__(self, exc_type, exc_value, traceback):
        self._local.current_transaction.pop()
        if self.rollback or exc_type is not None:
            self.connection.rollback()
        else:
            self.cursor.connection.commit()


execute = Transaction.execute
executemany = Transaction.executemany


def yield_from_cursor(cursor):
    while rows := cursor.fetchmany(1000):
        yield from rows


class ExecMany:
    """
    Helper class that can consume an iterator and feed the values
    yielded to a (returning) executemany statement.
    """

    def __init__(self, stm, values):
        self.stm = stm
        self.values = values

    def __iter__(self):
        # Create a dedicated cursor
        cursor = Transaction.current.connection.cursor()
        if Transaction.flavor == "sqlite":
            for vals in self.values:
                logger.debug(self.stm)
                cursor.execute(self.stm, vals)
                res = cursor.fetchone()
                yield res
        else:
            logger.debug(self.stm)
            cursor.executemany(self.stm, self.values, returning=True)
            while True:
                vals = cursor.fetchone()
                yield vals
                if not cursor.nextset():
                    break
